from .fair import (FairPredictor, inferred_attribute_builder, single_threshold, DataDict,
                   DeepFairPredictor, DeepDataDict)

__all__ = (FairPredictor, inferred_attribute_builder, single_threshold, DataDict, DeepFairPredictor,
           DeepDataDict)
